/*
	Engineering Test: Links Configuration
	Gordon Lee

*******************************************************************************

	@filename	cApplication.cpp

	@description
		Class for Application that handles the window and run loop

	@author		Gordon Lee
*/

// Includes
#include "cApplication.h"
#include "..\resource.h"
#include <io.h>
#include <fcntl.h>
#include <tchar.h>
#include <fstream>
#include <commdlg.h>
#include "cGraphics.h"
#include "cCamera.h"
#include "cInput.h"

namespace Test
{
////////////////////////////////////////////////////////////////////////////////
// Forward declarations
BOOL CALLBACK DlgProcNewChain(HWND, UINT, WPARAM, LPARAM);

/**
	Inline Proxy Win Proc

	@param HWND _hWnd
	@param UINT _msg
	@param WPARAM _wParam
	@param LPARAM _lParam

	@return LRESULT CALLBACK
*/
inline LRESULT CALLBACK ProxyWndProc(HWND _hWnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
	return g_pcApplication->AppWndProc(_hWnd, _msg, _wParam, _lParam);
}

/**
	Constructor
*/
cApplication::cApplication()
: m_winClassName(""), m_hInstance(NULL), m_hWnd(NULL), m_bConsoleOpen(false)
{
}

/**
	Deconstructor
*/
cApplication::~cApplication()
{
	cGraphics::Destroy();
	cInput::Destroy();

	if (m_chain)
		delete m_chain;

	// Handle application release
	UnregisterClass(m_winClassName.c_str(), m_hInstance);
}

/**
	Instance

	@return cApplication *
*/
cApplication *cApplication::Instance(void)
{
	static cApplication *s_application = new cApplication();
	return s_application;
}

/**
	Create

	@return cApplication *
*/
cApplication *cApplication::Create(void)
{
	return Instance();
}

/**
	Destroy
*/
void cApplication::Destroy(void)
{
	delete Instance();
}

/**
	Initialize

	@param HINSTANCE _hInstance
	@param std::string _title

	@return bool
*/
bool cApplication::Initialize(HINSTANCE _hInstance, std::string _title)
{
	HWND _hWnd;
	m_hInstance = _hInstance;

	m_winClassName = "Test";

	WNDCLASSEX winClass	    = { 0 };
	winClass.lpszClassName	= m_winClassName.c_str();
	winClass.cbSize			= sizeof(WNDCLASSEX);
	winClass.style			= CS_HREDRAW | CS_VREDRAW;
	winClass.lpfnWndProc	= ProxyWndProc;
	winClass.hInstance		= _hInstance;
	winClass.lpszMenuName	= (LPCTSTR)IDR_MENU1;
	winClass.hCursor		= LoadCursor(0, IDC_ARROW);
	winClass.hIcon			= LoadIcon(_hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
	winClass.hIconSm		= LoadIcon(winClass.hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
	winClass.hbrBackground	= (HBRUSH)COLOR_WINDOW;
	if (!RegisterClassEx(&winClass))
	{
		MessageBox(NULL, _T("Call to RegisterClassEx failed!"), _T("Reverse Words - cApplication::Initialize"), NULL);
		return false;
	}
	_hWnd = CreateWindowEx(0, m_winClassName.c_str(), _title.c_str(), WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
							CW_USEDEFAULT, CW_USEDEFAULT, SCREEN_WIDTH, SCREEN_HEIGHT, GetDesktopWindow(), 0, m_hInstance, NULL);
	if (!_hWnd)
	{
		MessageBox(NULL, _T("Call to CreateWindowEx failed!"), _T("Reverse Words - cApplication::Initialize"), NULL);
		return false;
	}
	m_hWnd = _hWnd;

	ShowWindow(m_hWnd, true);
	UpdateWindow(m_hWnd);

	// Graphics
	cGraphics::Create();
	g_pcGraphics->Initialize(m_hWnd);

	// Camera
	m_radius = 2.f;
	m_theta = -PI / 2.f;
	m_height = 0;
	g_pcGraphics->GetCamera()->SetPosition(D3DXVECTOR3(m_radius * cos(m_theta), m_height, m_radius * sin(m_theta)));
	g_pcGraphics->GetCamera()->SetLook(D3DXVECTOR3(-g_pcGraphics->GetCamera()->GetPosition().x, -m_height,
													-g_pcGraphics->GetCamera()->GetPosition().z));

	// Input
	cInput::Create();
	g_pcInput->Initialize();

	// Chain
	m_chain = new cChain(1.f, 1.f);
	g_pcGraphics->SetChain(m_chain);

	return true;
}

/**
	Open Console Window

	@return bool
*/
bool cApplication::OpenConsoleWindow(void)
{
	// Console window already open
	if (m_bConsoleOpen)
		return false;

	HANDLE _hStdHandle;
	int _hConHandle;
	FILE *_fp;

	if (!AllocConsole())
		return false;

	SetConsoleTitle("Links Configuration Debug Console");
	_hStdHandle = GetStdHandle(STD_OUTPUT_HANDLE);
	_hConHandle = _open_osfhandle((intptr_t)_hStdHandle, _O_TEXT);
	_fp = _fdopen(_hConHandle, "w");
	*stdout = *_fp;
	setvbuf(stdout, NULL, _IONBF, 0);
	std::ios::sync_with_stdio();

	m_bConsoleOpen = true;

	return true;
}

/**
	Close Console Window

	@return bool
*/
bool cApplication::CloseConsoleWindow(void)
{
	if (!FreeConsole())
		return false;

	m_bConsoleOpen = false;

	return true;
}

/**
	Run
*/
void cApplication::Run(void)
{
	// Message loop
	MSG msg = { 0 };

	while(msg.message != WM_QUIT)
	{
        if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		AppLoop();
	}
}

/**
	App Loop
*/
void cApplication::AppLoop(void)
{
	m_loopTimer.Start();

	//Input
	// Camera
	if (g_pcInput->KeyDown('W'))
	{
		m_radius = max(m_radius - 0.05f, 2.0f);
	}
	if (g_pcInput->KeyDown('S'))
	{
		m_radius += 0.05f;
	}
	if (g_pcInput->KeyDown('A'))
	{
		m_theta += -0.01f;
	}
	if (g_pcInput->KeyDown('D'))
	{
		m_theta += 0.01f;
	}
	if (g_pcInput->KeyDown('R'))
	{
		m_height += 0.05f;
	}
	if (g_pcInput->KeyDown('F'))
	{
		m_height += -0.05f;
	}

	// Spherical camera
	g_pcGraphics->GetCamera()->SetPosition(D3DXVECTOR3(m_radius * cos(m_theta), m_height, m_radius * sin(m_theta)));
	g_pcGraphics->GetCamera()->SetLook(D3DXVECTOR3(-g_pcGraphics->GetCamera()->GetPosition().x, -m_height,
													-g_pcGraphics->GetCamera()->GetPosition().z));

	// Link 1 start point moving
	if (g_pcInput->KeyDown('Y'))
		m_chain->m_pts[0].z += 0.005f;
	if (g_pcInput->KeyDown('H'))
		m_chain->m_pts[0].z += -0.005f;
	if (g_pcInput->KeyDown('J'))
		m_chain->m_pts[0].x += 0.005f;
	if (g_pcInput->KeyDown('G'))
		m_chain->m_pts[0].x += -0.005f;
	if (g_pcInput->KeyDown('I'))
		m_chain->m_pts[0].y += 0.005f;
	if (g_pcInput->KeyDown('K'))
		m_chain->m_pts[0].y += -0.005f;
	// Target moving
	if (g_pcInput->KeyDown(VK_NUMPAD8))
		m_chain->m_target.z += 0.005f;
	if (g_pcInput->KeyDown(VK_NUMPAD2))
		m_chain->m_target.z += -0.005f;
	if (g_pcInput->KeyDown(VK_NUMPAD6))
		m_chain->m_target.x += 0.005f;
	if (g_pcInput->KeyDown(VK_NUMPAD4))
		m_chain->m_target.x += -0.005f;
	if (g_pcInput->KeyDown(VK_NUMPAD7))
		m_chain->m_target.y += 0.005f;
	if (g_pcInput->KeyDown(VK_NUMPAD1))
		m_chain->m_target.y += -0.005f;
	// Solve
	if (g_pcInput->KeyPress(VK_RETURN))
	{
		if (m_bConsoleOpen)
			m_debugTimer.Start();

		m_chain->Solve();

		if (m_bConsoleOpen)
		{
			unsigned long _time = m_debugTimer.GetTime();
			printf("Link Configuration Algorithm Time: %d milliseconds\n\n", _time);
		}
	}
	g_pcInput->Update();

	// Graphics
	// UI text
	g_pcGraphics->DisplayText(0, 0,
								"Controls:\n\
								[W] - Camera move closer, [S] - Camera move back\n\
								[A] - Camera rotate left, [D] - Camera rotate right\n\
								[R] - Camera move up, [F] - Camera move down\n\
								[Y] - Link 1 start point +Z, [H] - Link 1 start point -Z\n\
								[J] - Link 1 start point +X, [G] - Link 1 start point -X\n\
								[I] - Link 1 start point +Y, [K] - Link 1 start point -Y\n\
								[NumPad8] - Target move +Z, [NumPad2] Target move -Z\n\
								[NumPad6] - Target move +X, [NumPad4] Target move -X\n\
								[NumPad7] - Target move +Y, [NumPad1] Target move -Y\n\
								[ENTER] - Solve and move to target\n\
								Red Line: X-axis, Green Line: Y-axis, Blue Line: Z-axis",
							  "Arial", D3DXCOLOR(255.f, 255.f, 255.f, 255.f), NULL);

	g_pcGraphics->Update();

	// Wait until timer reaches 1/60 of a second
	while (m_loopTimer.GetTime() < FRAMERATE_CAP);
}

/**
	App Wnd Proc

	@param HWND _hWnd
	@param UINT _msg
	@param WPARAM _wParam
	@param LPARAM _lParam

	@return LRESULT
*/
LRESULT cApplication::AppWndProc(HWND _hWnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
	switch (_msg)
	{
		case WM_CREATE:
			break;
		case WM_DESTROY:
			CloseConsoleWindow();
			PostQuitMessage(0);
			break;
		case WM_KEYDOWN:
			g_pcInput->ProcessKeyDown(_wParam);
			break;
		case WM_KEYUP:
			g_pcInput->ProcessKeyUp(_wParam);
			break;
		case WM_COMMAND:
			switch (_wParam)
			{
			// File/New Chain
			case ID_FILE_NEWCHAIN:
				DialogBox(m_hInstance, (LPCTSTR)IDD_DIALOG1, m_hWnd, (DLGPROC)DlgProcNewChain);
				if (m_bConsoleOpen)
					printf("New Chain; Link 1 Length: %f, Link 2 Length: %f\n", m_chain->m_links[0].m_length, m_chain->m_links[1].m_length);
				break;
			// File/Open Debug Console
			case ID_FILE_OPENDEBUGCONSOLE:
				OpenConsoleWindow();
				break;
			// File/Close Debug Console
			case ID_FILE_CLOSEDEBUGCONSOLE:
				CloseConsoleWindow();
				break;
			// File/Exit
			case ID_FILE_EXIT:
				CloseConsoleWindow();
				PostQuitMessage(0);
				break;
			// Help/How To
			case ID_HELP_HOWTO:
				MessageBox(NULL,
					_T("Menu:\nFile -> New Chain - Open a dialog box to create a new chain of inputted lengths\nFile -> Open Debug Console - Opens a console window for debug information such as algorithm time\nFile -> Close Debug Console - Closes the debug console window\n\nInstructions:\nThe camera uses spherical coordinates and moves using the keys: W, A, D, S, R, F.\nThe first link's starting point moves using the keys: Y, G, J, H, I, K.\nThe target moves using the keys: NumPad1, NumPad2, NumPad4, NumPad6, NumPad7, NumPad8\nPress Enter to solve the chain."),
					_T("How To"),
					MB_ICONINFORMATION);
				break;
			// Help/About
			case ID_HELP_ABOUT:
				MessageBox(NULL,
					_T("Link Configuration Engineer Test\nAuthor: Gordon Lee\nWritten in C++ and Microsoft Visual Studio 2008\nUses DirectX SDK (June 2010)\n\nI went with an IK closed-form solution method for the solution which finds the exact angles of the links to a configuration. I first rotate the target onto the XZ plane along the Z axis. Then, I solve for the rotated target on the XZ plane. Lastly, I rotate the target back to the original position, but now I can also use the same rotations on the links. I do not show the links if I cannot find a solution."),
					_T("About"),
					MB_ICONINFORMATION);
				break;
			default:
				break;
			}
		default:
			return DefWindowProc(_hWnd, _msg, _wParam, _lParam);
	}
	return DefWindowProc(_hWnd, _msg, _wParam, _lParam);
}

/**
	Dialog box function for new chain

	@param HWND _hDlg
	@param UINT _iMsg
	@param WPARAM _wParam
	@param LPARAM _lParam

	@return BOOL
*/
BOOL CALLBACK DlgProcNewChain(HWND _hDlg, UINT _iMsg, WPARAM _wParam, LPARAM _lParam)
{
	char _link1Length[BUF_LEN];
	char _link2Length[BUF_LEN];

	switch (_iMsg)
	{
	case WM_COMMAND:
		switch(LOWORD(_wParam))
		{
		case IDOK:
		{
			GetWindowText(GetDlgItem(_hDlg, IDC_EDIT1), _link1Length, BUF_LEN);
			GetWindowText(GetDlgItem(_hDlg, IDC_EDIT2), _link2Length, BUF_LEN);
			EndDialog(_hDlg, 0);
			
			delete g_pcApplication->GetChain();
			cChain *_chain = new cChain((float)atof(_link1Length), (float)atof(_link2Length));
			g_pcApplication->SetChain(_chain);
			break;
		}
		case IDCANCEL:
			EndDialog(_hDlg, -1);
			return false;
		default:
			break;
		}
	}
	return false;
}
////////////////////////////////////////////////////////////////////////////////
}
