/*
	Engineering Test: Links Configuration
	Gordon Lee

*******************************************************************************

	@filename	cChain.h

	@description
		Class for a chain of links solved by an IK solver

	@author		Gordon Lee
*/

#include "cChain.h"
#pragma comment(lib, "d3dx9.lib")
#include <math.h>
#include "cGraphics.h"
#include "GraphicsUtilities.h"
#include <stdio.h>

namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Assignment Constructor

	@param float _L1
	@param float _L2
*/
cChain::cChain(float _L1, float _L2)
{
	m_D3DDevice = g_pcGraphics->GetDevice();

	m_links[0].m_length = _L1;
	m_links[1].m_length = _L2;

	m_pts[0] = D3DXVECTOR3(0.f, 0.f, 0.f);
	m_pts[1] = D3DXVECTOR3(_L1, 0.f, 0.f);
	m_pts[2] = D3DXVECTOR3(m_pts[1].x + _L2, 0.f, 0.f);

	m_links[0].m_startPt = &m_pts[0];
	m_links[0].m_endPt = &m_pts[1];
	m_links[1].m_startPt = &m_pts[1];
	m_links[1].m_endPt = &m_pts[2];

	m_target = D3DXVECTOR3(0.5f, 0.f, 0.5f);
	D3DXCreateSphere(m_D3DDevice, 0.05f, 6, 6, &m_ptMesh, NULL);
}

/**
	Deconstructor
*/
cChain::~cChain()
{
	SAFE_RELEASE(m_ptMesh);
}

/**
	Draw
*/
void cChain::Draw(void)
{
	// Draw pts
	D3DXMATRIX _ptTrans;
	D3DXMatrixIdentity(&_ptTrans);
	D3DXMatrixTranslation(&_ptTrans, m_pts[0].x, m_pts[0].y, m_pts[0].z);
	m_D3DDevice->SetTransform(D3DTS_WORLD, &_ptTrans);
	m_ptMesh->DrawSubset(0);

	D3DXMatrixTranslation(&_ptTrans, m_pts[1].x, m_pts[1].y, m_pts[1].z);
	m_D3DDevice->SetTransform(D3DTS_WORLD, &_ptTrans);
	m_ptMesh->DrawSubset(0);

	D3DXMatrixTranslation(&_ptTrans, m_pts[2].x, m_pts[2].y, m_pts[2].z);
	m_D3DDevice->SetTransform(D3DTS_WORLD, &_ptTrans);
	m_ptMesh->DrawSubset(0);

	// Draw lines
	g_pcGraphics->DrawLine(m_pts[0], m_pts[1], D3DXCOLOR(255.f, 255.f, 255.f, 255.f));
	g_pcGraphics->DrawLine(m_pts[1], m_pts[2], D3DXCOLOR(255.f, 255.f, 255.f, 255.f));

	// Draw target
	D3DXMatrixTranslation(&_ptTrans, m_target.x, m_target.y, m_target.z);
	m_D3DDevice->SetTransform(D3DTS_WORLD, &_ptTrans);
	m_ptMesh->DrawSubset(0);
}

/**
	Solve
*/
void cChain::Solve(void)
{
	int _quadrant = 1;
	D3DXMATRIX _temp;
	D3DXMatrixIdentity(&_temp);
	float _angle = 0.f;

	m_target -= m_pts[0];

	// Rotate target to the XZ plane along Z axis
	if (m_target.x > 0.f)
	{
		D3DXVECTOR3 _v1(m_target.x, m_target.y, 0.f);
		D3DXVec3Normalize(&_v1, &_v1);
		D3DXVECTOR3 _v2(0.f, 1.f, 0.f);
		_angle = asin(D3DXVec3Dot(&_v1, &_v2));
		printf("First rotation angle onto XZ plane: %f\n", _angle);
		
		D3DXMatrixRotationZ(&_temp, -_angle);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
	}
	else if (m_target.x < 0.f)
	{
		D3DXVECTOR3 _v1(m_target.x, m_target.y, 0.f);
		D3DXVec3Normalize(&_v1, &_v1);
		D3DXVECTOR3 _v2(0.f, 1.f, 0.f);
		_angle = -asin(D3DXVec3Dot(&_v1, &_v2));
		printf("First rotation angle onto XZ plane: %f\n", _angle);
		
		D3DXMatrixRotationZ(&_temp, -_angle);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
	}
	else
	{
		_angle = PI / 2;
		printf("First rotation angle onto XZ plane: %f\n", _angle);
		
		D3DXMatrixRotationZ(&_temp, -_angle);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
	}

	// Calculate yaw rotation along Y axis on the XZ plane
	if (m_target.x < 0.f && m_target.z >= 0.f)
	{
		D3DXMatrixRotationY(&_temp, PI / 2);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		_quadrant = 2;
	}
	else if (m_target.x < 0.f && m_target.z < 0.f)
	{
		D3DXMatrixRotationY(&_temp, PI);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		_quadrant = 3;
	}
	else if (m_target.x >= 0.f && m_target.z < 0.f)
	{
		D3DXMatrixRotationY(&_temp, -PI / 2);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		_quadrant = 4;
	}

	// Closed-form solution of IK solving for the rotated target onto the XZ plane
	float _yaweq2 = ((m_target.x * m_target.x) + (m_target.z * m_target.z) -
					(m_links[0].m_length * m_links[0].m_length) - (m_links[1].m_length * m_links[1].m_length)) /
					(2.f * m_links[0].m_length * m_links[1].m_length);
	float _yaw2 = acos(_yaweq2);

	float _yaweq1 = (m_target.z * (m_links[0].m_length + m_links[1].m_length * cos(_yaw2)) -
					m_target.x * m_links[1].m_length * sin(_yaw2)) /
					((m_target.x * m_target.x) + (m_target.z * m_target.z));
	float _yaw1 = asin(_yaweq1);

	printf("Rotation of first link along Y axis: %f\n", _yaw1);
	printf("Rotation of second link from first along Y axis: %f\n", _yaw2);

	m_pts[1] = D3DXVECTOR3(m_links[0].m_length, 0.f, 0.f);
	m_pts[2] = D3DXVECTOR3(m_links[1].m_length, 0.f, 0.f);
	D3DXMatrixRotationY(&_temp, -_yaw1);
	D3DXVec3TransformCoord(&m_pts[1], &m_pts[1], &_temp);
	D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);
	D3DXMatrixRotationY(&_temp, -_yaw2);
	D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);
	m_pts[2] += m_pts[1];

	if (_quadrant == 2)
	{
		D3DXMatrixRotationY(&_temp, -PI / 2);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		D3DXVec3TransformCoord(&m_pts[1], &m_pts[1], &_temp);
		D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);
	}
	else if (_quadrant == 3)
	{
		D3DXMatrixRotationY(&_temp, -PI);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		D3DXVec3TransformCoord(&m_pts[1], &m_pts[1], &_temp);
		D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);
	}
	else if (_quadrant == 4)
	{
		D3DXMatrixRotationY(&_temp, PI / 2);
		D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
		D3DXVec3TransformCoord(&m_pts[1], &m_pts[1], &_temp);
		D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);
	}

	// Rotate everything back from the first rotation onto the XZ plane along Z axis
	D3DXMatrixRotationZ(&_temp, _angle);
	D3DXVec3TransformCoord(&m_target, &m_target, &_temp);
	D3DXVec3TransformCoord(&m_pts[1], &m_pts[1], &_temp);
	D3DXVec3TransformCoord(&m_pts[2], &m_pts[2], &_temp);

	m_target += m_pts[0];
	m_pts[1] += m_pts[0];
	m_pts[2] += m_pts[0];

	printf("Position of Link 1 Start: (%f, %f, %f)\n", m_pts[0].x, m_pts[0].y, m_pts[0].z);
	printf("Position of Link 1 End and Link 2 Start: (%f, %f, %f)\n", m_pts[1].x, m_pts[1].y, m_pts[1].z);
	printf("Position of Link 2 End: (%f, %f, %f)\n", m_pts[2].x, m_pts[2].y, m_pts[2].z);
}
////////////////////////////////////////////////////////////////////////////////
}
