/*
	Engineering Test: Links Configuration
	Gordon Lee

*******************************************************************************

	@filename	cGraphics.cpp

	@description
		Class for a graphics engine using DirectX

	@author		Gordon Lee
*/

// Includes
#include "cGraphics.h"
#include "GraphicsUtilities.h"
#include "cCamera.h"
#include "cGraphicsRenderer.h"
#include "cFontManager.h"

// Namespace
namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Constructor
*/
cGraphics::cGraphics()
{
}

/**
	Deconstructor
*/
cGraphics::~cGraphics()
{
	SAFE_DELETE(m_camera);
	SAFE_DELETE(m_renderer);
	SAFE_DELETE(m_fontManager);
}

/**
	Instance

	@return cGraphics *
*/
cGraphics *cGraphics::Instance(void)
{
	static cGraphics *s_graphics = new cGraphics();
	return s_graphics;
}

/**
	Create

	@return cGraphics *
*/
cGraphics *cGraphics::Create(void)
{
	return Instance();
}

/**
	Destroy
*/
void cGraphics::Destroy(void)
{
	delete Instance();
}

/**
	Initialize

	@param HWND _hWnd

	@return bool
*/
bool cGraphics::Initialize(HWND _hWnd)
{
	m_camera = new cCamera();
	m_renderer = new cGraphicsRenderer();
	m_fontManager = new cFontManager();

	if (!m_renderer->Initialize(_hWnd))
		return false;

	if (!m_fontManager->Initialize(m_renderer->GetDevice()))
		return false;

	return true;
}

/**
	Update
*/
void cGraphics::Update(void)
{
	if (!m_renderer->IsDeviceLost())
	{
		m_renderer->BuildProjection();
		SetupCamera();
		m_renderer->SetView(m_camera->GetView());

		BeginRendering(D3DCOLOR_XRGB(0, 0, 50));
			DrawOrigin();
			m_chain->Draw();
			RenderText();
		EndRendering();
	}
}

/**
	Get Device

	@return LPDIRECT3DDEVICE9
*/
LPDIRECT3DDEVICE9 cGraphics::GetDevice(void)
{
	return m_renderer->GetDevice();
}

/**
	Draw Text

	@param unsigned _width
	@param unsigned _height
	@param std::string _text
	@param std::string _fontName
	@param D3DXCOLOR _color
	@param DWORD _flags
*/
void cGraphics::DisplayText(unsigned _width, unsigned _height, std::string _text, std::string _fontName, D3DXCOLOR _color, DWORD _flags)
{
	sRenderPacketText *_textPacket = new sRenderPacketText;
	_textPacket->m_font = m_fontManager->GetFont(_fontName);
	_textPacket->m_text = _text;
	_textPacket->m_rect.left = _width;
	_textPacket->m_rect.top = _height;
	_textPacket->m_rect.right = _width;
	_textPacket->m_rect.bottom = _height;
	_textPacket->m_color = _color;
	_textPacket->m_flags = _flags;
	
	m_fontManager->PushTextPacket(_textPacket);
}

/**
	Render Text
*/
void cGraphics::RenderText(void)
{
	TextStack *_stack = m_fontManager->GetTextStack();
	while (!_stack->empty())
	{
		m_renderer->RenderText(_stack->top());
		delete _stack->top();
		_stack->pop();
	}
}

/**
	Setup Camera
*/
void cGraphics::SetupCamera(void)
{
	m_camera->Orthonormalize();
	m_camera->BuildView();
}

/**
	Begin Rendering

	@param D3DCOLOR _color
*/
void cGraphics::BeginRendering(D3DCOLOR _color)
{
	m_renderer->BeginRendering(_color);
}

/**
	Draw Line

	@param D3DXVECTOR3 _p0
	@param D3DXVECTOR3 _p1
	@param D3DXCOLOR _color
*/
void cGraphics::DrawLine(D3DXVECTOR3 _p0, D3DXVECTOR3 _p1, D3DXCOLOR _color)
{
	m_renderer->RenderLine(_p0, _p1, _color);
}

/**
	End Rendering
*/
void cGraphics::EndRendering(void)
{
	m_renderer->EndRendering();
}

/**
	Draw Origin
*/
void cGraphics::DrawOrigin(void)
{
	// X
	DrawLine(D3DXVECTOR3(0.f, 0.f, 0.f), D3DXVECTOR3(1.f, 0.f, 0.f), D3DXCOLOR(255.f, 0.f, 0.f, 255.f));
	// Y
	DrawLine(D3DXVECTOR3(0.f, 0.f, 0.f), D3DXVECTOR3(0.f, 1.f, 0.f), D3DXCOLOR(0.f, 255.f, 0.f, 255.f));
	// Z
	DrawLine(D3DXVECTOR3(0.f, 0.f, 0.f), D3DXVECTOR3(0.f, 0.f, 1.f), D3DXCOLOR(0.f, 0.f, 255.f, 255.f));
}
////////////////////////////////////////////////////////////////////////////////
}
