/*
	Engineering Test: Links Configuration
	Gordon Lee

*******************************************************************************

	@filename	cGraphicsRenderer.h

	@description
		Class for the rendering of the graphics engine

	@author		Gordon Lee
*/

// Includes
#include "cGraphicsRenderer.h"
#pragma comment(lib, "d3d9.lib")

// Namespace
namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Constructor
*/
cGraphicsRenderer::cGraphicsRenderer()
: m_D3DObject(NULL), m_D3DDevice(NULL), m_D3DLine(NULL)
{
}

/**
	Deconstructor
*/
cGraphicsRenderer::~cGraphicsRenderer()
{
	SAFE_RELEASE(m_D3DObject);

	SAFE_RELEASE(m_D3DDevice);

	SAFE_RELEASE(m_D3DLine);
}

/**
	Initialize

	@param HWND _hWnd

	@return bool
*/
bool cGraphicsRenderer::Initialize(HWND _hWnd)
{
	// Create IDirect3D9 object
	m_D3DObject = Direct3DCreate9(D3D_SDK_VERSION);

	CheckDeviceCaps();

	m_D3DPP.BackBufferWidth				= 1024;
	m_D3DPP.BackBufferHeight			= 768;
	m_D3DPP.BackBufferFormat			= D3DFMT_X8R8G8B8;
	m_D3DPP.BackBufferCount           	= 1;
    m_D3DPP.MultiSampleType           	= D3DMULTISAMPLE_NONE;
    m_D3DPP.MultiSampleQuality        	= 0;
    m_D3DPP.SwapEffect                	= D3DSWAPEFFECT_DISCARD;
    m_D3DPP.hDeviceWindow             	= _hWnd;
    m_D3DPP.Windowed                  	= true;
    m_D3DPP.EnableAutoDepthStencil    	= true;
    m_D3DPP.AutoDepthStencilFormat    	= D3DFMT_D24S8;
    m_D3DPP.Flags                     	= 0;
    m_D3DPP.FullScreen_RefreshRateInHz	= D3DPRESENT_RATE_DEFAULT;
    m_D3DPP.PresentationInterval      	= D3DPRESENT_INTERVAL_IMMEDIATE;

	// Create the device
	D3DHR(m_D3DObject->CreateDevice(D3DADAPTER_DEFAULT,
									D3DDEVTYPE_HAL,
									m_D3DPP.hDeviceWindow,
									m_D3DDeviceCaps.m_capsFlags,
									&m_D3DPP,
									&m_D3DDevice));

	// Set render states
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE));			// turn off lighting
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_ZENABLE, TRUE));			// turn on z-buffer
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE));	
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_AMBIENT, D3DCOLOR_XRGB(250, 250, 250)));

	// D3DXLine
	D3DHR(D3DXCreateLine(m_D3DDevice, &m_D3DLine));

	return true;
}

/**
	Check Device Caps
*/
void cGraphicsRenderer::CheckDeviceCaps(void)
{
	// Check backbuffer formats
	D3DDISPLAYMODE _mode;
	m_D3DObject->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &_mode);
    D3DHR(m_D3DObject->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
          _mode.Format, _mode.Format, true));
     
    D3DHR(m_D3DObject->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
          D3DFMT_X8R8G8B8, D3DFMT_X8R8G8B8, false));

	// Check the device capabilities for vertex processing
	D3DCAPS9 _caps;
	D3DHR(m_D3DObject->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &_caps));

	if (_caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
		m_D3DDeviceCaps.m_capsFlags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		m_D3DDeviceCaps.m_capsFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	// Check the device for shader support
    if (_caps.VertexShaderVersion < D3DVS_VERSION(3,0))
		m_D3DDeviceCaps.m_vertexShader3_0Support = false;

    if (_caps.PixelShaderVersion < D3DPS_VERSION(3,0))
		m_D3DDeviceCaps.m_pixelShader3_0Support = false;

	// See how many shader constants are available
	m_D3DDeviceCaps.m_maxConstants = _caps.MaxVertexShaderConst;

	// Ensure enough render targets
	m_D3DDeviceCaps.m_maxRenderTargets = _caps.NumSimultaneousRTs;
}

/**
	Is Device Lost

	@return bool
*/
bool cGraphicsRenderer::IsDeviceLost(void)
{
	if (!m_D3DDevice)
		return false;

	HRESULT hr = m_D3DDevice->TestCooperativeLevel();

	if (hr == D3DERR_DEVICELOST)
	{
		Sleep(20);
		return true;
	}
	else if (hr == D3DERR_DRIVERINTERNALERROR)
		return true;
	else if (hr == D3DERR_DEVICENOTRESET)
		return false;
	else
		return false;
}

/**
	Build Projection
*/
void cGraphicsRenderer::BuildProjection(void)
{
	float _width = (float)m_D3DPP.BackBufferWidth;
	float _height = (float)m_D3DPP.BackBufferHeight;
	D3DXMatrixPerspectiveFovLH(&m_proj, D3DX_PI * 0.25f, _width / _height, 0.1f, 1000.f);
	m_D3DDevice->SetTransform(D3DTS_PROJECTION, &m_proj);
}

/**
	Begin rendering

	@param D3DCOLOR _color
*/
void cGraphicsRenderer::BeginRendering(D3DCOLOR _color)
{
	m_D3DDevice->SetTransform(D3DTS_VIEW, &m_view);
	m_D3DDevice->BeginScene();
	m_D3DDevice->Clear(0, 0, D3DCLEAR_STENCIL | D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, _color, 1.f, 0);

	// Render states
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME));
	D3DHR(m_D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW));
}

/**
	End rendering
*/
void cGraphicsRenderer::EndRendering(void)
{
	m_D3DDevice->EndScene();
	m_D3DDevice->Present(0, 0, 0, 0);
}

/**
	Render Line

	@param D3DXVECTOR3 _p0
	@param D3DXVECTOR3 _p1
	@param D3DXCOLOR _color
*/
void cGraphicsRenderer::RenderLine(D3DXVECTOR3 _p0, D3DXVECTOR3 _p1, D3DXCOLOR _color)
{
	D3DXVECTOR3 _points[2] = { _p0, _p1 };
	D3DXMATRIX _viewProj = m_view * m_proj;
	D3DHR(m_D3DLine->DrawTransform(_points, 2, &_viewProj, _color));
}

/**
	Render Text

	@param sRenderPacketText *_packet
*/
void cGraphicsRenderer::RenderText(sRenderPacketText *_packet)
{
	D3DHR(_packet->m_font->DrawText(NULL, _packet->m_text.c_str(), -1, &_packet->m_rect, _packet->m_flags | DT_NOCLIP, _packet->m_color));
}
////////////////////////////////////////////////////////////////////////////////
}
